from datetime import datetime, date, timedelta
from dateutil import rrule
from dateutil.relativedelta import relativedelta

from django.test import TestCase, override_settings
from django.utils.timezone import get_default_timezone, make_aware
from django.conf import settings
from eventtools.models import REPEAT_MAX

from .models import Event, Occurrence


class EventToolsTestCase(TestCase):

    def setUp(self):
        self.christmas = Event.objects.create(title='Christmas')
        Occurrence.objects.create(
            event=self.christmas,
            start=datetime(2000, 12, 25, 7, 0),
            end=datetime(2000, 12, 25, 22, 0),
            repeat="RRULE:FREQ=YEARLY")

        self.weekends = Event.objects.create(title='Weekends 9-10am')
        # Saturday
        Occurrence.objects.create(
            event=self.weekends,
            start=datetime(2015, 1, 3, 9, 0),
            end=datetime(2015, 1, 3, 10, 0),
            repeat="RRULE:FREQ=WEEKLY")
        # Sunday
        Occurrence.objects.create(
            event=self.weekends,
            start=datetime(2015, 1, 4, 9, 0),
            end=datetime(2015, 1, 4, 10, 0),
            repeat="RRULE:FREQ=WEEKLY")

        self.daily = Event.objects.create(title='Daily 7am')
        Occurrence.objects.create(
            event=self.daily,
            start=datetime(2015, 1, 1, 7, 0),
            end=None,
            repeat="RRULE:FREQ=DAILY")

        self.past = Event.objects.create(title='Past event')
        Occurrence.objects.create(
            event=self.past,
            start=datetime(2014, 1, 1, 7, 0),
            end=datetime(2014, 1, 1, 8, 0))

        self.future = Event.objects.create(title='Future event')
        Occurrence.objects.create(
            event=self.future,
            start=datetime(2016, 1, 1, 7, 0),
            end=datetime(2016, 1, 1, 8, 0))

        self.monthly = Event.objects.create(title='Monthly until Dec 2017')
        Occurrence.objects.create(
            event=self.monthly,
            start=datetime(2016, 1, 1, 7, 0),
            end=datetime(2016, 1, 1, 8, 0),
            repeat="RRULE:FREQ=MONTHLY",
            repeat_until=date(2017, 12, 31))

        # fake "today" so tests always work
        self.today = date(2015, 6, 1)
        self.first_of_year = date(2015, 1, 1)
        self.last_of_year = date(2015, 12, 31)

    def test_single_occurrence(self):
        occ = self.christmas.occurrence_set.get()

        # using date() arguments
        dates = list(occ.all_occurrences(
            from_date=date(2015, 12, 1),
            to_date=date(2015, 12, 31),))
        self.assertEqual(len(dates), 1)

        # check it works as expected when from/to equal the occurrence date
        dates = list(occ.all_occurrences(
            from_date=date(2015, 12, 25),
            to_date=date(2015, 12, 25), ))
        self.assertEqual(len(dates), 1)

        # using datetime() arguments
        dates = list(occ.all_occurrences(
            from_date=datetime(2015, 12, 25, 6, 0, 0),
            to_date=datetime(2015, 12, 25, 23, 0, 0), ))
        self.assertEqual(len(dates), 1)

        # using tz-aware datetime() arguments, if appropriate
        if settings.USE_TZ:
            tz = get_default_timezone()
            dates = list(occ.all_occurrences(
                from_date=datetime(2015, 12, 25, 6, 0, 0, 0, tz),
                to_date=datetime(2015, 12, 25, 23, 0, 0, 0, tz), ))
            self.assertEqual(len(dates), 1)

        # date range intersecting with occurrence time
        dates = list(occ.all_occurrences(
            from_date=datetime(2015, 12, 25, 10, 0, 0),
            to_date=datetime(2015, 12, 25, 23, 0, 0), ))
        self.assertEqual(len(dates), 1)
        dates = list(occ.all_occurrences(
            from_date=datetime(2015, 12, 25, 6, 0, 0),
            to_date=datetime(2015, 12, 25, 10, 0, 0), ))
        self.assertEqual(len(dates), 1)

        # date range within occurrence time
        dates = list(occ.all_occurrences(
            from_date=datetime(2015, 12, 25, 12, 0, 0),
            to_date=datetime(2015, 12, 25, 13, 0, 0), ))
        self.assertEqual(len(dates), 1)

        # date range outside occurrence time
        dates = list(occ.all_occurrences(
            from_date=datetime(2015, 12, 24, 12, 0, 0),
            to_date=datetime(2015, 12, 26, 13, 0, 0), ))
        self.assertEqual(len(dates), 1)

        # date range before occurrence time
        dates = list(occ.all_occurrences(
            from_date=datetime(2015, 12, 24, 12, 0, 0),
            to_date=datetime(2015, 12, 24, 13, 0, 0), ))
        self.assertEqual(len(dates), 0)

        # date range after occurrence time
        dates = list(occ.all_occurrences(
            from_date=datetime(2015, 12, 25, 23, 0, 0),
            to_date=datetime(2015, 12, 25, 23, 30, 0), ))
        self.assertEqual(len(dates), 0)

        # check next_occurence method for non-repeating occurrences
        occ = self.past.occurrence_set.get() \
                  .next_occurrence(from_date=self.today)
        self.assertEqual(occ, None)

        occ = self.future.occurrence_set.get() \
                  .next_occurrence(from_date=self.today)
        self.assertEqual(occ[0].timetuple()[:5],
                         datetime(2016, 1, 1, 7, 0).timetuple()[:5])

        # and for repeating
        occ = self.daily.occurrence_set.get() \
                  .next_occurrence(from_date=self.today)
        self.assertEqual(occ[0].date(), self.today)

        # test next_occurrence for querysets
        occ = self.daily.occurrence_set.all() \
                  .next_occurrence(from_date=self.today)
        self.assertEqual(occ[0].date(), self.today)

    @override_settings(USE_TZ=True)
    def test_single_occurrence_tz(self):
        self.test_single_occurrence()

    def test_occurrence_qs(self):
        events = [self.christmas, self.past, self.future]
        occs = Occurrence.objects.filter(event__in=events)

        # two christmases and the future event
        dates = list(occs.all_occurrences(
            from_date=date(2015, 1, 1),
            to_date=date(2016, 12, 31),))
        self.assertEqual(len(dates), 3)

        # one christmas and the past event
        dates = list(occs.all_occurrences(
            from_date=date(2014, 1, 1),
            to_date=date(2014, 12, 31),))
        self.assertEqual(len(dates), 2)

        # test queryset filtering
        qs = occs.for_period(from_date=date(2015, 1, 1), exact=True)
        self.assertEqual(qs.count(), 2)

        qs = occs.for_period(to_date=date(2010, 1, 1), exact=True)
        self.assertEqual(qs.get().event, self.christmas)

        qs = occs.for_period(from_date=date(2017, 1, 1),
                             to_date=date(2017, 12, 31),
                             exact=True)
        self.assertEqual(qs.get().event, self.christmas)

    @override_settings(USE_TZ=True)
    def test_occurrence_qs_tz(self):
        self.test_occurrence_qs()

    def test_single_event(self):
        # one christmas per year
        for i in range(0, 10):
            count = len(list(self.christmas.all_occurrences(
                from_date=self.first_of_year + relativedelta(years=i),
                to_date=self.first_of_year + relativedelta(years=i + 1))))
            self.assertEqual(count, 1)

        # but none in the first half of the year
        count = len(list(self.christmas.all_occurrences(
            from_date=self.first_of_year + relativedelta(years=1),
            to_date=self.first_of_year + relativedelta(months=6))))
        self.assertEqual(count, 0)

        # check the daily event happens on some arbitrary dates
        for days in (10, 30, 50, 80, 100):
            from_date = self.first_of_year + timedelta(days)
            count = len(list(self.daily.all_occurrences(
                from_date=from_date,
                to_date=from_date
            )))
            self.assertEqual(count, 1)

        # check the the weekend event occurs as expected in a series of 2 day
        # periods
        for days in range(1, 50):
            from_date = self.first_of_year + timedelta(days)

            if from_date.weekday() == 5:
                expected = 2  # whole weekend
            elif from_date.weekday() in (4, 6):
                expected = 1  # one weekend day
            else:
                expected = 0  # no weekend days

            occs = list(self.weekends.all_occurrences(
                from_date=from_date,
                to_date=from_date + timedelta(1)
            ))
            self.assertEqual(len(occs), expected)

    @override_settings(USE_TZ=True)
    def test_single_event_tz(self):
        self.test_single_event()

    def test_event_queryset(self):
        # one christmas per year
        christmas_qs = Event.objects.filter(pk=self.christmas.pk)
        for i in range(0, 10):
            occs = list(christmas_qs.all_occurrences(
                from_date=self.first_of_year + relativedelta(years=i),
                to_date=self.first_of_year + relativedelta(years=i + 1)))
            self.assertEqual(len(occs), 1)

        # but none in the first half of the year
        occs = list(christmas_qs.all_occurrences(
            from_date=self.first_of_year + relativedelta(years=1),
            to_date=self.first_of_year + relativedelta(months=6)))
        self.assertEqual(len(occs), 0)

        def sorted_events(events):
            return sorted(events, key=lambda obj: obj.pk)

        def expected(for_date):
            """Return ids of events expected to occur on a given date. """

            events = [self.daily]
            if for_date.month == 12 and for_date.day == 25:
                events.append(self.christmas)
            if for_date.weekday() in (5, 6):
                events.append(self.weekends)
            return sorted_events(events)

        # check the number of events for some arbitrary dates
        for days in (8, 16, 24, 32, 40, 48, 56):
            from_date = self.first_of_year + timedelta(days)
            qs = Event.objects.for_period(
                from_date=from_date,
                to_date=from_date,
                exact=True,
            ).distinct()

            events = sorted_events(list(qs))
            self.assertEqual(events, expected(from_date))

        # test queryset filtering
        events = Event.objects.filter(
            pk__in=(self.christmas.pk, self.future.pk, self.past.pk))

        qs = events.for_period(from_date=date(2015, 1, 1), exact=True)
        self.assertEqual(qs.count(), 2)

        qs = events.for_period(to_date=date(2010, 1, 1), exact=True)
        self.assertEqual(qs.get(), self.christmas)

        qs = events.for_period(from_date=date(2017, 1, 1),
                               to_date=date(2017, 12, 31),
                               exact=True)
        self.assertEqual(qs.get(), self.christmas)

    @override_settings(USE_TZ=True)
    def test_event_queryset_tz(self):
        self.test_event_queryset()

    def test_occurrence_data(self):
        occ = self.christmas.occurrence_set.get()
        self.assertEqual(occ.next_occurrence()[2], occ.occurrence_data)

    def test_repeat_until(self):
        # check repeating event when to_date is less than repeat_until
        occs = self.monthly.all_occurrences(from_date=date(2016, 4, 1),
                                            to_date=date(2016, 4, 30))
        self.assertEqual(len(list(occs)), 1)

    def test_occurrence_limit(self):
        test_objs = [
            self.daily,
            Event.objects.filter(pk=self.daily.pk),
            self.daily.occurrence_set.all(),
            self.daily.occurrence_set.get(),
        ]
        for obj in test_objs:
            self.assertEqual(len(list(obj.all_occurrences(limit=20))), 20)
            self.assertEqual(len(list(obj.all_occurrences())), REPEAT_MAX)

    def test_non_repeating_intersection(self):
        occ = self.past.occurrence_set.get()

        dates = list(occ.all_occurrences(
            from_date=datetime(2014, 1, 1, 7, 30),
            to_date=datetime(2014, 1, 1, 8, 30)))
        self.assertEqual(len(dates), 1)
        dates = list(occ.all_occurrences(
            from_date=datetime(2014, 1, 1, 6, 30),
            to_date=datetime(2014, 1, 1, 7, 30)))
        self.assertEqual(len(dates), 1)

    def test_integer_rules_can_be_migrated(self):
        yearly = Occurrence.objects.create(
            event=self.christmas,
            start=datetime(2000, 12, 25, 7, 0),
            end=datetime(2000, 12, 25, 22, 0),
            repeat=rrule.YEARLY)
        monthly = Occurrence.objects.create(
            event=self.past,
            start=datetime(2014, 1, 1, 7, 0),
            end=datetime(2014, 1, 1, 8, 0),
            repeat=rrule.MONTHLY)
        weekly = Occurrence.objects.create(
            event=self.weekends,
            start=datetime(2015, 1, 4, 9, 0),
            end=datetime(2015, 1, 4, 10, 0),
            repeat=rrule.WEEKLY)
        daily = Occurrence.objects.create(
            event=self.daily,
            start=datetime(2015, 1, 1, 7, 0),
            end=datetime(2015, 1, 1, 8, 0),
            repeat=rrule.DAILY)

        Occurrence.objects.migrate_integer_repeat()
        yearly.refresh_from_db()
        self.assertEqual(yearly.repeat, 'RRULE:FREQ=YEARLY')
        monthly.refresh_from_db()
        self.assertEqual(monthly.repeat, 'RRULE:FREQ=MONTHLY')
        weekly.refresh_from_db()
        self.assertEqual(weekly.repeat, 'RRULE:FREQ=WEEKLY')
        daily.refresh_from_db()
        self.assertEqual(daily.repeat, 'RRULE:FREQ=DAILY')

    def test_queryset_filtering(self):
        event = Event.objects.create(title='One off')
        Occurrence.objects.create(
            event=event,
            start=datetime(2000, 1, 1, 7, 0),
            end=datetime(2000, 1, 1, 8, 0))
        events = Event.objects.filter(pk=event.pk)
        occs = event.occurrence_set.all()

        # for_period should be accurate in this simple case
        self.assertEqual(
            occs.for_period(date(2000, 1, 1)).count(),
            events.for_period(date(2000, 1, 1)).count(),
            1)

        self.assertEqual(
            occs.for_period(date(2001, 1, 1)).count(),
            events.for_period(date(2001, 1, 1)).count(),
            0)

        self.assertEqual(
            occs.for_period(date(1999, 1, 1), date(1999, 1, 2)).count(),
            events.for_period(date(1999, 1, 1), date(1999, 1, 2)).count(),
            0)

        self.assertEqual(
            occs.for_period(date(1999, 1, 1), date(2001, 1, 2)).count(),
            events.for_period(date(1999, 1, 1), date(2001, 1, 2)).count(),
            0)

    @override_settings(USE_TZ=True, TIME_ZONE='America/New_York')
    def test_dst_boundary(self):
        # Check that event start times are consistent across daylight saving
        # changes - on an EST5EDT system, daylight saving ends on 5/11/2016
        event = Event.objects.create(title='Test')
        start = make_aware(datetime(2016, 11, 5, 10, 0))
        Occurrence.objects.create(event=event, start=start,
                                  repeat="RRULE:FREQ=WEEKLY")

        occs = list(event.all_occurrences(from_date=start, limit=2))
        self.assertEqual(occs[0][0], start)
        self.assertEqual(occs[1][0], make_aware(datetime(2016, 11, 12, 10, 0)))

    @override_settings(USE_TZ=True, TIME_ZONE='Pacific/Auckland')
    def test_dst_boundary_nz(self):
        # NZ DST commences on 25/9/2016
        event = Event.objects.create(title='Test')
        start = make_aware(datetime(2016, 9, 20, 10, 0))
        Occurrence.objects.create(event=event, start=start,
                                  repeat="RRULE:FREQ=WEEKLY")

        occs = list(event.all_occurrences(from_date=start.date(), limit=2))
        self.assertEqual(occs[0][0], start)
        self.assertEqual(occs[1][0], make_aware(datetime(2016, 9, 27, 10, 0)))

    def test_sort_by_next(self):
        qs = Event.objects.filter(pk__in=[self.christmas.pk, self.weekends.pk])

        # Christmas 2015 fell on a Friday
        christmas_first = qs.sort_by_next(from_date=date(2015, 12, 24))
        self.assertEqual(christmas_first, [self.christmas, self.weekends])

        weekend_first = qs.sort_by_next(from_date=date(2015, 12, 20))
        self.assertEqual(weekend_first, [self.weekends, self.christmas])



# -*- coding: utf-8 -*-

from dateutil import rrule
from datetime import date, datetime, timedelta

from django.conf import settings
from django.db import models
from django.db.models import Q, Case, When, Value
from django.core.exceptions import ValidationError
from django.utils.timezone import make_aware, is_naive, make_naive, is_aware
from django.utils.encoding import python_2_unicode_compatible


# set EVENTTOOLS_REPEAT_CHOICES = None to make this a plain textfield
REPEAT_CHOICES = getattr(settings, 'EVENTTOOLS_REPEAT_CHOICES', (
    ("RRULE:FREQ=DAILY", 'Daily'),
    ("RRULE:FREQ=WEEKLY", 'Weekly'),
    ("RRULE:FREQ=MONTHLY", 'Monthly'),
    ("RRULE:FREQ=YEARLY", 'Yearly'),
))
REPEAT_MAX = 200


def max_future_date():
    return datetime(date.today().year + 10, 1, 1, 0, 0)


def first_item(gen):
    try:
        return next(gen)
    except StopIteration:
        return None


def default_aware(dt):
    """Convert a naive datetime argument to a tz-aware datetime, if tz support
       is enabled. """

    if settings.USE_TZ and is_naive(dt):
        return make_aware(dt)

    # if timezone support disabled, assume only naive datetimes are used
    return dt


def default_naive(dt):
    """Convert an aware datetime argument to naive, if tz support
       is enabled. """

    if settings.USE_TZ and is_aware(dt):
        return make_naive(dt)

    # if timezone support disabled, assume only naive datetimes are used
    return dt


def as_datetime(d, end=False):
    """Normalise a date/datetime argument to a datetime for use in filters

    If a date is passed, it will be converted to a datetime with the time set
    to 0:00, or 23:59:59 if end is True."""

    if type(d) is date:
        date_args = tuple(d.timetuple())[:3]
        if end:
            time_args = (23, 59, 59)
        else:
            time_args = (0, 0, 0)
        new_value = datetime(*(date_args + time_args))
        return default_aware(new_value)
    # otherwise assume it's a datetime
    return default_aware(d)


def combine_occurrences(generators, limit):
    """Merge the occurrences in two or more generators, in date order.

       Returns a generator. """

    count = 0
    grouped = []
    for gen in generators:
        try:
            next_date = next(gen)
        except StopIteration:
            pass
        else:
            grouped.append({'generator': gen, 'next': next_date})

    while limit is None or count < limit:
        # all generators must have finished if there are no groups
        if not len(grouped):
            return

        # work out which generator will yield the earliest date (based on
        # start - end is ignored)
        next_group = None
        for group in grouped:
            if not next_group or group['next'][0] < next_group['next'][0]:
                next_group = group

        # yield the next (start, end) pair, with occurrence data
        yield next_group['next']
        count += 1

        # update the group's next item, so we don't keep yielding the same date
        try:
            next_group['next'] = next(next_group['generator'])
        except StopIteration:
            # remove the group if there's none left
            grouped.remove(next_group)


def filter_invalid(approx_qs, from_date, to_date):
    """Filter out any results from the queryset which do not have an occurrence
       within the given range. """

    # work out what to exclude based on occurrences
    exclude_pks = []
    for obj in approx_qs:
        if not obj.next_occurrence(from_date=from_date, to_date=to_date):
            exclude_pks.append(obj.pk)

    # and then apply the filtering to the queryset itself
    return approx_qs.exclude(pk__in=exclude_pks)


class OccurrenceMixin(object):
    """Class mixin providing common occurrence-related functionality. """

    def all_occurrences(self, from_date=None, to_date=None):
        raise NotImplementedError()

    def next_occurrence(self, from_date=None, to_date=None):
        """Return next occurrence as a (start, end) tuple for this instance,
           between from_date and to_date, taking repetition into account. """
        if not from_date:
            from_date = datetime.now()
        return first_item(
            self.all_occurrences(from_date=from_date, to_date=to_date))

    def first_occurrence(self):
        """Return first occurrence as a (start, end) tuple for this instance.
        """
        return first_item(self.all_occurrences())


class BaseQuerySet(models.QuerySet, OccurrenceMixin):
    """Base QuerySet for models which have occurrences. """

    def for_period(self, from_date=None, to_date=None, exact=False):
        # subclasses should implement this
        raise NotImplementedError()

    def sort_by_next(self, from_date=None):
        """Sort the queryset by next_occurrence.

        Note that this method necessarily returns a list, not a queryset. """

        def sort_key(obj):
            occ = obj.next_occurrence(from_date=from_date)
            return occ[0] if occ else None
        return sorted([e for e in self if sort_key(e)], key=sort_key)

    def all_occurrences(self, from_date=None, to_date=None, limit=None):
        """Return a generator yielding a (start, end) tuple for all occurrence
           dates in the queryset, taking repetition into account, up to a
           maximum limit if specified. """

        # winnow out events which are definitely invalid
        qs = self.for_period(from_date, to_date)

        return combine_occurrences(
            (obj.all_occurrences(from_date, to_date) for obj in qs), limit)


class BaseModel(models.Model, OccurrenceMixin):
    """Abstract model providing common occurrence-related functionality. """

    class Meta:
        abstract = True


class EventQuerySet(BaseQuerySet):
    """QuerySet for BaseEvent subclasses. """

    def for_period(self, from_date=None, to_date=None, exact=False):
        """Filter by the given dates, returning a queryset of Occurrence
           instances with occurrences falling within the range.

           Due to uncertainty with repetitions, from_date filtering is only an
           approximation. If exact results are needed, pass exact=True - this
           will use occurrences to exclude invalid results, but may be very
           slow, especially for large querysets. """

        filtered_qs = self

        # to_date filtering is accurate
        if to_date:
            to_date = as_datetime(to_date, True)
            filtered_qs = filtered_qs.filter(
                Q(occurrence__start__lte=to_date)).distinct()

        if from_date:
            # but from_date isn't, due to uncertainty with repetitions, so
            # just winnow down as much as possible via queryset filtering
            from_date = as_datetime(from_date)
            filtered_qs = filtered_qs.filter(
                Q(occurrence__end__isnull=False,
                  occurrence__end__gte=from_date) |
                Q(occurrence__start__gte=from_date) |
                (~Q(occurrence__repeat='') &
                 (Q(occurrence__repeat_until__gte=from_date) |
                  Q(occurrence__repeat_until__isnull=True)))).distinct()

            # filter out invalid results if requested
            if exact:
                filtered_qs = filter_invalid(filtered_qs, from_date, to_date)

        return filtered_qs


class EventManager(models.Manager.from_queryset(EventQuerySet)):
    use_for_related_fields = True


class BaseEvent(BaseModel):
    """Abstract model providing occurrence-related methods for events.

       Subclasses should have a related BaseOccurrence subclass. """

    objects = EventManager()

    def all_occurrences(self, from_date=None, to_date=None, limit=None):
        """Return a generator yielding a (start, end) tuple for all dates
           for this event, taking repetition into account. """

        return self.occurrence_set.all_occurrences(from_date, to_date,
                                                   limit=limit)

    class Meta:
        abstract = True


class OccurrenceQuerySet(BaseQuerySet):
    """QuerySet for BaseOccurrence subclasses. """

    def for_period(self, from_date=None, to_date=None, exact=False):
        """Filter by the given dates, returning a queryset of Occurrence
           instances with occurrences falling within the range.

           Due to uncertainty with repetitions, from_date filtering is only an
           approximation. If exact results are needed, pass exact=True - this
           will use occurrences to exclude invalid results, but may be very
           slow, especially for large querysets. """

        filtered_qs = self

        # to_date filtering is accurate
        if to_date:
            to_date = as_datetime(to_date, True)
            filtered_qs = filtered_qs.filter(Q(start__lte=to_date)).distinct()

        if from_date:
            # but from_date isn't, due to uncertainty with repetitions, so
            # just winnow down as much as possible via queryset filtering
            from_date = as_datetime(from_date)
            filtered_qs = filtered_qs.filter(
                Q(end__isnull=False, end__gte=from_date) |
                Q(start__gte=from_date) |
                (~Q(repeat='') &
                 (Q(repeat_until__gte=from_date) |
                  Q(repeat_until__isnull=True)))).distinct()

            # filter out invalid results if requested
            if exact:
                filtered_qs = filter_invalid(filtered_qs, from_date, to_date)

        return filtered_qs


class OccurrenceManager(models.Manager.from_queryset(OccurrenceQuerySet)):
    use_for_related_fields = True

    def migrate_integer_repeat(self):
        self.update(repeat=Case(
            When(repeat=rrule.YEARLY,
                 then=Value("RRULE:FREQ=YEARLY")),
            When(repeat=rrule.MONTHLY,
                 then=Value("RRULE:FREQ=MONTHLY")),
            When(repeat=rrule.WEEKLY,
                 then=Value("RRULE:FREQ=WEEKLY")),
            When(repeat=rrule.DAILY,
                 then=Value("RRULE:FREQ=DAILY")),
            default=Value(""),
        ))


class ChoiceTextField(models.TextField):
    """Textfield which uses a Select widget if it has choices specified. """

    def formfield(self, **kwargs):
        if self.choices:
            # this overrides the TextField's preference for a Textarea widget,
            # allowing the ModelForm to decide which field to use
            kwargs['widget'] = None
        return super(ChoiceTextField, self).formfield(**kwargs)


@python_2_unicode_compatible
class BaseOccurrence(BaseModel):
    """Abstract model providing occurrence-related methods for occurrences.

       Subclasses will usually have a ForeignKey pointing to a BaseEvent
       subclass. """

    start = models.DateTimeField(db_index=True)
    end = models.DateTimeField(db_index=True, null=True, blank=True)

    repeat = ChoiceTextField(choices=REPEAT_CHOICES, default='', blank=True)
    repeat_until = models.DateField(null=True, blank=True)

    def clean(self):
        if self.start and self.end and self.start >= self.end:
            msg = u"End must be after start"
            raise ValidationError(msg)

        if self.repeat_until and self.repeat is None:
            msg = u"Select a repeat interval, or remove the " \
                  u"'repeat until' date"
            raise ValidationError(msg)

        if self.start and self.repeat_until and \
           self.repeat_until < self.start.date():
            msg = u"'Repeat until' cannot be before the first occurrence"
            raise ValidationError(msg)

    objects = OccurrenceManager()

    def all_occurrences(self, from_date=None, to_date=None, limit=REPEAT_MAX):
        """Return a generator yielding a (start, end) tuple for all dates
           for this occurrence, taking repetition into account. """

        if not self.start:
            return

        from_date = from_date and as_datetime(from_date)
        to_date = to_date and as_datetime(to_date, True)

        if not self.repeat:
            if (not from_date or self.start >= from_date or
                (self.end and self.end >= from_date)) and \
               (not to_date or self.start <= to_date):
                yield (self.start, self.end, self.occurrence_data)
        else:
            delta = (self.end - self.start) if self.end else timedelta(0)
            repeater = self.get_repeater()

            # start from the first occurrence at the earliest
            if not from_date or from_date < self.start:
                from_date = self.start

            # look until the last occurrence, up to an arbitrary maximum date
            if self.repeat_until and (
                    not to_date or
                    as_datetime(self.repeat_until, True) < to_date):
                to_date = as_datetime(self.repeat_until, True)
            elif not to_date:
                to_date = default_aware(max_future_date())

            # start is used for the filter, so modify from_date to take the
            # occurrence length into account
            from_date -= delta

            # always send naive datetimes to the repeater
            repeater = repeater.between(default_naive(from_date),
                                        default_naive(to_date), inc=True)

            count = 0
            for occ_start in repeater:
                count += 1
                if count > limit:
                    return

                # make naive results aware
                occ_start = default_aware(occ_start)
                yield (occ_start, occ_start + delta, self.occurrence_data)

    def get_repeater(self):
        # Timings to get all_occurrences() for a set of 2500 Occurrence objects
        # with rrule.DAILY repeat
        # Without method call (inline repeat)
        # CPU times: user 53.4 s, sys: 76 ms, total: 53.4 s
        # Wall time: 55.8 s

        # With method call
        # CPU times: user 53.5 s, sys: 100 ms, total: 53.6 s
        # Wall time: 56 s
        # The subclassing benefit seems much larger than the performance hit

        return rrule.rrulestr(self.repeat, dtstart=default_naive(self.start))

    @property
    def occurrence_data(self):
        return self

    class Meta:
        ordering = ('start', 'end')
        abstract = True

    def __str__(self):
        return u"%s" % (self.start)



from django.db import models
from django.utils.encoding import python_2_unicode_compatible

from eventtools.models import BaseEvent, BaseOccurrence


@python_2_unicode_compatible
class Event(BaseEvent):
    title = models.CharField(max_length=100)

    def __str__(self):
        return self.title


class Occurrence(BaseOccurrence):
    event = models.ForeignKey(Event)
